import os

from nucliadb_sdk.v2.exceptions import NotFoundError

from nuclia.sdk.kb import NucliaKB
from nuclia.sdk.upload import NucliaUpload

path = f"{os.path.dirname(__file__)}/assets/conversation.json"


def test_conversation(testing_config):
    nkb = NucliaKB()
    try:
        resource = nkb.get_resource_by_slug(slug="conversation1")
        nkb.delete(rid=resource.id)
    except NotFoundError:
        pass

    res = nkb.list()
    assert res is None

    nu = NucliaUpload()
    nu.conversation(path=path)

    resource = nkb.get_resource_by_slug(slug="conversation1")
    assert resource
    assert resource.data.conversations["conversation1"]



from typing import Optional

import requests

from nuclia import REGIONAL
from nuclia.exceptions import NuaAPIException
from nuclia.lib.nua_responses import Sentence

SENTENCE_PREDICT = "/api/v1/predict/sentence"


class NuaClient:
    def __init__(self, region: str, account: str, token: str):
        self.region = region
        self.account = account
        self.token = token
        self.url = REGIONAL.format(region=region).strip("/")
        self.headers = {"X-STF-NUAKEY": f"Bearer {token}"}

    def sentence_predict(self, text: str, model: Optional[str] = None) -> Sentence:
        resp = requests.get(
            f"{self.url}{SENTENCE_PREDICT}?text={text}&model={model}",
            headers=self.headers,
        )
        if resp.status_code == 200:
            return Sentence.parse_obj(resp.json())
        else:
            raise NuaAPIException()



from typing import List

from nucliadb_models import Message
from pydantic import BaseModel


class ConversationWrapper(BaseModel):
    slug: str
    messages: List[Message]


class Conversations(BaseModel):
    conversations: List[ConversationWrapper]



import os


def yes_no(question: str) -> bool:
    if os.environ.get("TESTING") == "True":
        return True
    yes_choices = ["yes", "y"]
    no_choices = ["no", "n"]
    value = None

    while True:
        user_input = input(f"{question} (yes/no): ")
        if user_input.lower() in yes_choices:
            value = True
            break
        elif user_input.lower() in no_choices:
            value = False
            break
        else:
            print("Type yes or no")
            continue
    return value



from nucliadb_models.resource import Resource, ResourceList

from nuclia.data import get_auth
from nuclia.decorators import kb
from nuclia.lib.kb import NucliaDBClient
from nuclia.sdk.auth import NucliaAuth
from nuclia.sdk.search import NucliaSearch
from nuclia.sdk.upload import NucliaUpload


class NucliaKB:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    def __init__(self):
        self.upload = NucliaUpload()
        self.search = NucliaSearch()

    @kb
    def list(self, *, ndb: NucliaDBClient):
        data: ResourceList = ndb.ndb.list_resources(kbid=ndb.kbid)
        for resource in data.resources:
            print(f"{resource.id} {resource.icon:30} {resource.title}")

    @kb
    def get_resource_by_slug(self, *, ndb: NucliaDBClient, slug: str) -> Resource:
        return ndb.ndb.get_resource_by_slug(
            kbid=ndb.kbid, slug=slug, query_params={"show": "values"}
        )

    @kb
    def delete(self, *, ndb: NucliaDBClient, rid: str):
        ndb.ndb.delete_resource(kbid=ndb.kbid, rid=rid)



import base64
import json
import readline  # noqa
import webbrowser
from typing import Dict, List, Optional, Tuple

import requests

from nuclia import BASE, get_global_url
from nuclia.cli.utils import yes_no
from nuclia.config import Account, Config, KnowledgeBox, Zone
from nuclia.exceptions import NeedUserToken, UserTokenExpired

USER = f"{BASE}/api/v1/user/welcome"
MEMBER = f"{BASE}/api/v1/user"
ACCOUNTS = f"{BASE}/api/v1/accounts"
ZONES = f"{BASE}/api/v1/zones"
LIST_KBS = BASE + "/api/v1/account/{account}/kbs"
VERIFY_NUA = "/api/authorizer/info"


class NucliaAuth:
    _inner_config: Optional[Config] = None

    @property
    def _config(self) -> Config:
        if self._inner_config is None:
            from nuclia.data import get_config

            self._inner_config = get_config()
        return self._inner_config

    def show(self):
        if self._config.default:
            print("Default Knowledge Box")
            print("=====================")
            print()
            print(self._config.get_kb(self._config.default.kbid))
            print()

        if self._config.token:
            print("User Auth")
            print("=========")
            print()
            self._show_user()
            print()

        if len(self._config.kbs_token):
            print("Knowledgebox Auth")
            print("=================")
            print()
            for kb in self._config.kbs_token:
                print(kb)
            print()

        if len(self._config.nuas_token):
            print("NUA Key")
            print("=======")
            print()
            for nua in self._config.nuas_token:
                print(nua)

    def kb(self, url: str, token: str):
        url = url.strip("/")
        kbid, title = self.validate_kb(url, token)
        if kbid:
            print("Validated")
            try:
                next(
                    filter(
                        lambda x: x.id == kbid,
                        self._config.kbs if self._config.kbs is not None else [],
                    )
                )
                if yes_no(f"Want to replace actual KB {kbid} configuration") is False:
                    print("Cancelling operation")
                    return

            except StopIteration:
                # Not found
                pass

            self._config.set_kb_token(url=url, token=token, title=title, kbid=kbid)
        else:
            print("Invalid service token")

    def nua(self, region: str, token: str) -> Optional[str]:
        client_id, account_type, account = self.validate_nua(token, region)
        if account is not None and client_id is not None:
            print("Validated")
            self._config.set_nua_token(
                client_id=client_id,
                account_type=account_type,
                account=account,
                region=region,
                token=token,
            )
            return client_id
        else:
            print("Invalid service token")
            return None

    def validate_nua(
        self, token: str, region: Optional[str] = None
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        # Validate the code is ok
        url = get_global_url(VERIFY_NUA)
        resp = requests.get(
            url,
            headers={"x-nuclia-nuakey": f"Bearer {token}"},
        )
        if resp.status_code == 200:
            data = resp.json().get("user")
            return data.get("user_id"), data.get("account_type"), data.get("account_id")
        else:
            return None, None, None

    def validate_kb(self, url: str, token: str) -> Tuple[Optional[str], Optional[str]]:
        # Validate the code is ok
        resp = requests.get(
            url,
            headers={"X-Nuclia-Serviceaccount": f"Bearer {token}"},
        )
        if resp.status_code == 200:
            data = resp.json()
            return data.get("uuid"), data.get("config", {}).get("title")
        else:
            return None, None

    def _show_user(self):
        resp = self.get_user(MEMBER)
        print(f"User: {resp.get('name')} <{resp.get('email')}>")
        print(f"Type: {resp.get('type')}")

    def login(self):
        """
        Lets redirect the user to the UI to capture a token
        """
        if self._validate_user_token():
            print("Logged in!")
            self._show_user()
            return

        webbrowser.open(get_global_url("/redirect?display=token&ident={ident}"))
        code = input("Follow the browser flow and copy the token and paste it here:")
        print("Checking...")
        self.set_user_token(code)

    def set_user_token(self, code: str):
        if self._validate_user_token(code):
            self._config.set_user_token(code)
            print("Auth completed!")
            self.post_login()
        else:
            print("Invalid token auth not completed")

    def _extract_account(self, token: str) -> str:
        base64url = token.split(".")[1]
        data = json.loads(
            base64.urlsafe_b64decode(base64url + "=" * (4 - len(base64url) % 4))
        )
        return data.get("jti")

    def _validate_user_token(self, code: Optional[str] = None) -> bool:
        # Validate the code is ok
        if code is None:
            code = self._config.token
        resp = requests.get(
            USER,
            headers={"Authorization": f"Bearer {code}"},
        )
        if resp.status_code == 200:
            return True
        else:
            return False

    def post_login(self):
        self.accounts()
        self.zones()

    def post_user(self, path: str, payload: Dict[str, str]):
        if not self._config.token:
            raise NeedUserToken()
        resp = requests.post(
            path,
            headers={"Authorization": f"Bearer {self._config.token}"},
            data=payload,
        )
        if resp.status_code == 201:
            return resp.json()
        elif resp.status_code == 403:
            raise UserTokenExpired()

    def get_user(self, path: str):
        if not self._config.token:
            raise NeedUserToken()
        resp = requests.get(
            path,
            headers={"Authorization": f"Bearer {self._config.token}"},
        )
        if resp.status_code == 200:
            return resp.json()
        elif resp.status_code == 403:
            raise UserTokenExpired()

    def accounts(self) -> List[Account]:
        accounts = self.get_user(ACCOUNTS)
        result = []
        self._config.accounts = []
        for account in accounts:
            account_obj = Account.parse_obj(account)
            result.append(account_obj)
            self._config.accounts.append(account_obj)
        self._config.save()
        return result

    def zones(self) -> List[Zone]:
        zones = self.get_user(ZONES)
        if self._config.accounts is None:
            self._config.accounts = []
        self._config.zones = []
        result = []
        for zone in zones:
            zone_obj = Zone.parse_obj(zone)
            result.append(zone_obj)
            self._config.zones.append(zone_obj)
        self._config.save()
        return result

    def kbs(self, account: str):
        path = LIST_KBS.format(account=account)
        try:
            kbs = self.get_user(path)
        except UserTokenExpired:
            return []
        result = []
        zones = self.zones()
        region = {zone.id: zone.slug for zone in zones}
        for kb in kbs:
            zone = region[kb["zone"]]
            url = f"https://{zone}.nuclia.cloud/api/v1/kb/{kb['id']}"
            kb_obj = KnowledgeBox(
                url=url, id=kb["id"], title=kb["title"], account=account, region=zone
            )
            result.append(kb_obj)
        return result



from __future__ import annotations

import mimetypes
import os
import sys
from datetime import datetime
from pathlib import Path
from typing import Optional
from uuid import uuid4

import requests
from nucliadb_models.metadata import Extra, Origin
from nucliadb_models.text import TextFormat
from tqdm import tqdm

from nuclia.data import get_auth
from nuclia.decorators import kb
from nuclia.lib.conversations import Conversations
from nuclia.lib.kb import NucliaDBClient
from nuclia.sdk.auth import NucliaAuth


class NucliaUpload:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @kb
    def file(
        self,
        *,
        ndb: NucliaDBClient,
        path: str,
        rid: Optional[str] = None,
        field: Optional[str] = None,
    ):
        """Option to upload a file from filesystem to a Nuclia KnowledgeBox"""
        filename = path.split("/")[-1]
        size = os.path.getsize(path)
        mimetype_result = mimetypes.guess_type(path)
        if None in mimetype_result:
            mimetype = "application/octet-stream"
        else:
            mimetype = "/".join(mimetype_result)  # type: ignore
        with open(path, "rb") as upload_file:
            upload_url = ndb.start_tus_upload(
                rid=rid,
                field=field,
                size=size,
                filename=filename,
                content_type=mimetype,
            )

            offset = 0
            for _ in tqdm(range((size // 524288) + 1)):
                chunk = upload_file.read(524288)
                offset = ndb.patch_tus_upload(
                    upload_url=upload_url, data=chunk, offset=offset
                )

    @kb
    def conversation(self, *, ndb: NucliaDBClient, path: str):
        """Option to upload a conversation from filesystem to a Nuclia KnowledgeBox"""
        conversations = Conversations.parse_file(path)
        if conversations.conversations is None:
            return

        for conversation in conversations.conversations:
            if conversation.messages is None:
                continue
            rid = conversation.slug if conversation.slug is not None else uuid4().hex
            ndb.ndb.create_resource(
                kbid=ndb.kbid,
                slug=rid,
                icon="application/conversation",
                conversations={
                    rid: {
                        "messages": [
                            {
                                "who": message.who
                                if message.who is not None
                                else uuid4().hex,
                                "to": [x for x in message.to]
                                if message.to is not None
                                else [],
                                "ident": message.ident
                                if message.ident is not None
                                else uuid4().hex,
                                "timestamp": message.timestamp
                                if message.timestamp is not None
                                else datetime.now().isoformat(),
                                "content": {
                                    "text": message.content.text,
                                    "format": message.content.format
                                    if message.content.format is not None
                                    else "PLAIN",
                                },
                            }
                            for message in conversation.messages
                        ]
                    }
                },
            )

    @kb
    def text(
        self,
        *,
        ndb: NucliaDBClient,
        format: TextFormat = TextFormat.PLAIN,
        path: Optional[str] = None,
        stdin: Optional[bool] = False,
        slug: Optional[str] = None,
        origin: Optional[Origin] = None,
        extra: Optional[Extra] = None,
    ):
        """Option to upload a text from filesystem to a Nuclia KnowledgeBox"""
        if path is None and not stdin:
            raise ValueError("Either path or stdin must be provided")
        if path:
            text = Path(path).resolve().open().read()
        else:
            text = sys.stdin.read()
        rid = slug if slug is not None else uuid4().hex
        icon = "text/plain"
        if format == "HTML":
            icon = "text/html"
        elif format == "MARKDOWN":
            icon = "text/markdown"
        elif format == "RST":
            icon = "text/x-rst"
        ndb.ndb.create_resource(
            kbid=ndb.kbid,
            slug=rid,
            icon=icon,
            origin=origin,
            extra=extra,
            texts={
                rid: {
                    "body": text,
                    "format": format,
                }
            },
        )

    @kb
    def remote(
        self,
        *,
        ndb: NucliaDBClient,
        origin: str,
        rid: Optional[str] = None,
        field: Optional[str] = "file",
    ):
        """Option to upload a remote url to a Nuclia KnowledgeBox"""
        with requests.get(origin, stream=True) as r:
            filename = origin.split("/")[-1]
            size_str = r.headers.get("Content-Length")
            if size_str is None:
                size_str = "-1"
            size = int(size_str)
            upload_url = ndb.start_tus_upload(
                rid=rid,
                field=field,
                size=size,
                filename=filename,
                content_type=r.headers.get("Content-Type", "application/octet-stream"),
            )
            offset = 0
            for _ in tqdm(range((size // 524288) + 1)):
                chunk = r.raw.read(524288)
                offset = ndb.patch_tus_upload(upload_url, chunk, offset)



import fire  # type: ignore

from nuclia.data import get_auth
from nuclia.sdk.accounts import NucliaAccounts
from nuclia.sdk.kb import NucliaKB
from nuclia.sdk.kbs import NucliaKBS
from nuclia.sdk.nua import NucliaNUA
from nuclia.sdk.nuas import NucliaNUAS
from nuclia.sdk.zones import NucliaZones


class NucliaCLI(object):
    def __init__(self):
        self.auth = get_auth()
        self.accounts = NucliaAccounts()
        self.zones = NucliaZones()
        self.kbs = self.knowledgeboxes = NucliaKBS()
        self.kb = self.knowledgebox = NucliaKB()
        self.nuas = NucliaNUAS()
        self.nua = NucliaNUA()


def run():
    fire.Fire(NucliaCLI)



from typing import Dict, Optional

from nuclia import BASE
from nuclia.config import retrieve
from nuclia.data import get_auth
from nuclia.decorators import accounts
from nuclia.sdk.auth import NucliaAuth

ADD_KB = BASE + "/api/v1/account/{account}/kbs"


class NucliaKBS:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @accounts
    def list(self, account: Optional[str] = None):
        if account is None:
            result = []
            accounts = (
                self._auth._config.accounts
                if self._auth._config.accounts is not None
                else []
            )
            for account_obj in accounts:
                if account_obj.slug is not None:
                    result.extend(self._auth.kbs(account_obj.slug))
            self._auth._config.kbs = result
            self._auth._config.save()

            # List the Knowledge Boxes configured as Service Token
            result.extend(
                self._auth._config.kbs_token
                if self._auth._config.kbs_token is not None
                else []
            )

            return result
        else:
            return self._auth.kbs(account)

    def add(
        self,
        account: str,
        slug: str,
        anonymization: Optional[str] = None,
        description: Optional[str] = None,
        learning_configuration: Optional[Dict[str, str]] = None,
        sentence_embedder: Optional[str] = None,
        title: Optional[str] = None,
        zone: Optional[str] = None,
    ):
        # path = ADD_KB.format(account=account)
        raise NotImplementedError()
        # return self._auth.post_user(path, payload)

    def default(self, kb: str):
        kbs = self._auth._config.kbs if self._auth._config.kbs is not None else []
        kb_obj = retrieve(kbs, kb)
        if kb_obj is None:
            kbs = (
                self._auth._config.kbs_token
                if self._auth._config.kbs_token is not None
                else []
            )
            kb_obj = retrieve(kbs, kb)

        if kb_obj is None:
            raise KeyError("Knowledge Box not found")
        self._auth._config.set_default_kb(kb_obj.id)
        self._auth._config.save()



from typing import Optional

from nuclia.data import get_auth
from nuclia.decorators import nua
from nuclia.lib.nua import NuaClient
from nuclia.lib.nua_responses import Sentence
from nuclia.sdk.auth import NucliaAuth


class NucliaPredict:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @nua
    def sentence(
        self, nc: NuaClient, text: str, model: Optional[str] = None
    ) -> Sentence:
        return nc.sentence_predict(text, model)
